@@ -9,7 +9,8 @@ from .modulation import get_modulation_type_from_order, get_scale_coef_constella
     get_nearest_constellation_points_unscaled
 from .metrics import get_ber_by_points, get_ber_by_points_ultimate, get_energy, get_average_power, get_evm_ultimate
 
-from ssfm_gpu.propagation import propagate_manakov, propagate_schrodinger, dispersion_compensation_manakov
+from ssfm_gpu.propagation import propagate_manakov, propagate_manakov_backward, \
+    propagate_schrodinger, dispersion_compensation_manakov
 
 # Channel parameters
 
@@ -119,7 +120,7 @@ def full_line_model_default():
     return full_line_model(channel, wdm)
 
 
-def full_line_model(channel, wdm, bits_x=None, bits_y=None, points_x=None, points_y=None, verbose=0):
+def full_line_model(channel, wdm, bits=None, points=None, verbose=0):
     """
     Simulates a full optical transmission line, including generation of a wavelength division multiplexed (WDM)
     signal with one WDM channel, propagation through a specified channel, and detection at the receiver.
@@ -127,10 +128,8 @@ def full_line_model(channel, wdm, bits_x=None, bits_y=None, points_x=None, point
     Args:
         channel: object, channel through which the WDM signal is passed
         wdm: dict, contains information about the WDM signal
-        bits_x: int, number of bits in the x component of the signal (optional)
-        bits_y: int, number of bits in the y component of the signal (optional)
-        points_x: array, points of the x component of the signal (optional)
-        points_y: array, points of the y component of the signal (optional)
+        bits_x: tuple, number of bits in the x (and y) component of the signal (optional)
+        points_x: tuple, points of the x (and y) component of the signal (optional)
         verbose: int: level of system messages. 0 -- nothing, 1 -- +metrics, 2 -- +time, 3 -- +everything (optional)
 
     Returns:
@@ -157,7 +156,7 @@ def full_line_model(channel, wdm, bits_x=None, bits_y=None, points_x=None, point
     sample_freq = int(wdm['symb_freq'] * wdm['upsampling'])
     dt = 1. / sample_freq
 
-    signal_x, signal_y, wdm_info = generate_wdm(wdm)
+    signal_x, signal_y, wdm_info = generate_wdm(wdm, bits=bits, points=points)
     # generate_wdm is for multichannel wdm
     # for only one channel we have to take [0] element in list
     # that will correspond to desired values
@@ -243,8 +242,7 @@ def full_line_model(channel, wdm, bits_x=None, bits_y=None, points_x=None, point
 
 
 # @execution_time
-def full_line_model_wdm(channel, wdm, bits_x=None, bits_y=None,
-                        points_x=None, points_y=None, channels_type='all', verbose=0):
+def full_line_model_wdm(channel, wdm, bits=None, points=None, channels_type='all', verbose=0):
     """
         Simulates a full optical transmission line, including generation of a wavelength division multiplexed (WDM)
         signal with multiple WDM channels, propagation through a specified channel, and detection at the receiver.
@@ -252,10 +250,8 @@ def full_line_model_wdm(channel, wdm, bits_x=None, bits_y=None,
         Args:
             channel: object, channel through which the WDM signal is passed
             wdm: dict, contains information about the WDM signal
-            bits_x: int, number of bits in the x component of the signal (optional)
-            bits_y: int, number of bits in the y component of the signal (optional)
-            points_x: array, points of the x component of the signal (optional)
-            points_y: array, points of the y component of the signal (optional)
+            bits: tuple, number of bits in the x (and y) component of the signal (optional)
+            points: tuple, points of the x (and y) component of the signal (optional)
             channels_type: str: 'all' -- calculate metrics for all channels, 'middle' -- calculate only for middle (optional)
             verbose: int: level of system messages. 0 -- nothing, 1 -- +metrics, 2 -- +time, 3 -- +everything (optional)
 
@@ -279,7 +275,7 @@ def full_line_model_wdm(channel, wdm, bits_x=None, bits_y=None,
 
     dt = 1. / wdm['sample_freq']
 
-    signal_x, signal_y, wdm_info = generate_wdm(wdm)
+    signal_x, signal_y, wdm_info = generate_wdm(wdm, bits=bits, points=points)
 
     points_x_orig = wdm_info['points_x']
     points_y_orig = wdm_info['points_y']
@@ -499,4 +495,154 @@ def full_line_model_optimise(channel, wdm, points_orig_x, points_orig_y, ft_tx_f
             'q_y': q_y
         }
 
+    return result
+
+
+def full_line_model_back_to_back(channel, wdm, bits_x=None, bits_y=None, points_x=None, points_y=None, verbose=0):
+    """
+    Simulates a full optical transmission line, including generation of a wavelength division multiplexed (WDM)
+    signal with one WDM channel, propagation through a specified channel, backward propagation,
+    and detection at the receiver.
+
+    Args:
+        channel: object, channel through which the WDM signal is passed
+        wdm: dict, contains information about the WDM signal
+        bits_x: int, number of bits in the x component of the signal (optional)
+        bits_y: int, number of bits in the y component of the signal (optional)
+        points_x: array, points of the x component of the signal (optional)
+        points_y: array, points of the y component of the signal (optional)
+        verbose: int: level of system messages. 0 -- nothing, 1 -- +metrics, 2 -- +time, 3 -- +everything (optional)
+
+    Returns:
+        dict: containing the points and BER and Q-value of the signal in the x and y component
+
+        - 'points_x' -- an array of the points of the x component of the signal after processing
+        - 'points_x_orig' -- an array of the original points of the x component of the signal
+        - 'points_x_shifted' -- an array of the points of the x component of the signal after shifting
+        - 'points_x_found' -- an array of the nearest constellation points of the x component of the signal
+        - 'points_y' -- an array of the points of the y component of the signal after processing
+        - 'points_y_orig' -- an array of the original points of the y component of the signal
+        - 'points_y_shifted' -- an array of the points of the y component of the signal after shifting
+        - 'points_y_found' -- an array of the nearest constellation points of the y component of the signal
+        - 'ber_x' -- the bit error rate of the x component of the signal
+        - 'ber_y' -- the bit error rate of the y component of the signal
+        - 'q_x' -- the Q-value of the x component of the signal
+        - 'q_y' -- the Q-value of the y component of the signal
+
+    """
+
+    if wdm['n_channels'] > 1:
+        raise Exception('Wrong number of WDM channels for [full_line_model] use [full_line_model_wdm]')
+
+    sample_freq = int(wdm['symb_freq'] * wdm['upsampling'])
+    dt = 1. / sample_freq
+
+    signal_x, signal_y, wdm_info = generate_wdm(wdm)
+    # generate_wdm is for multichannel wdm
+    # for only one channel we have to take [0] element in list
+    # that will correspond to desired values
+    points_x_orig = wdm_info['points_x'][0]
+    points_y_orig = wdm_info['points_y'][0]
+    ft_filter_values = wdm_info['ft_filter_values_x'][0]
+    np_signal = len(signal_x)
+
+    e_signal_x = get_energy(signal_x, dt * np_signal)
+    e_signal_y = get_energy(signal_y, dt * np_signal)
+    p_signal_x = get_average_power(signal_x, dt)
+    p_signal_y = get_average_power(signal_y, dt)
+    p_signal_correct = dbm_to_mw(wdm['p_ave_dbm']) / 1000 / wdm['n_polarisations'] * wdm['n_channels']
+    print("Average signal power (x / y): "
+          "%1.7f / %1.7f (has to be close to %1.7f)" % (
+          p_signal_x, p_signal_y, p_signal_correct)) if verbose >= 3 else ...
+
+    start_time = datetime.now()
+    signal_x, signal_y = propagate_manakov(channel, signal_x, signal_y, wdm['sample_freq'])
+    print("forward propagation took", (datetime.now() - start_time).total_seconds() * 1000, "ms") if verbose >= 2 else ...
+
+    e_signal_x_prop = get_energy(signal_x, dt * np_signal)
+    e_signal_y_prop = get_energy(signal_y, dt * np_signal)
+
+    if verbose >= 3:
+        print("Signal energy before propagation (x / y):", e_signal_x, e_signal_y)
+        print("Signal energy after propagation (x / y):", e_signal_x_prop, e_signal_y_prop)
+        print("Signal energy difference (x / y):",
+              np.absolute(e_signal_x - e_signal_x_prop),
+              np.absolute(e_signal_y - e_signal_y_prop))
+
+    channel_back = channel.copy()
+    channel_back['z_span'] = -channel['z_span']
+
+    # [propagate_manakov_backward] is the proper function for backpropagation
+    # which properly handle attenuation of the signal
+    # if you still wand to use [propagate_manakov] then you have to set alpha = -alpha for channel parameters
+    start_time = datetime.now()
+    signal_x, signal_y = propagate_manakov_backward(channel_back, signal_x, signal_y, wdm['sample_freq'])
+    print("backward propagation took", (datetime.now() - start_time).total_seconds() * 1000,
+          "ms") if verbose >= 2 else ...
+
+    e_signal_x_prop_back = get_energy(signal_x, dt * np_signal)
+    e_signal_y_prop_back = get_energy(signal_y, dt * np_signal)
+
+    if verbose >= 3:
+        print("Signal energy before propagation (x / y):", e_signal_x, e_signal_y)
+        print("Signal energy after propagation (x / y):", e_signal_x_prop, e_signal_y_prop)
+        print("Signal energy after backward propagation (x / y):", e_signal_x_prop_back, e_signal_y_prop_back)
+        print("Signal energy difference (x / y):",
+              np.absolute(e_signal_x - e_signal_x_prop),
+              np.absolute(e_signal_y - e_signal_y_prop),
+              np.absolute(e_signal_x - e_signal_x_prop_back),
+              np.absolute(e_signal_y - e_signal_y_prop_back)
+              )
+
+    samples_x, samples_y = receiver(signal_x, signal_y, ft_filter_values, wdm['downsampling_rate'])
+    # for back-to-back we don't need dispersion compensation
+    # samples_x, samples_y = dispersion_compensation_manakov(channel, samples_x, samples_y, dt * wdm['downsampling_rate'])
+
+    sample_step = int(wdm['upsampling'] / wdm['downsampling_rate'])
+    points_x = samples_x[::sample_step].numpy()
+    points_y = samples_y[::sample_step].numpy()
+
+    nl_shift_x = nonlinear_shift(points_x, points_x_orig)
+    points_x_shifted = points_x * nl_shift_x
+
+    nl_shift_y = nonlinear_shift(points_y, points_y_orig)
+    points_y_shifted = points_y * nl_shift_y
+
+    mod_type = get_modulation_type_from_order(wdm['m_order'])
+    scale_constellation = get_scale_coef_constellation(mod_type) / np.sqrt(wdm['p_ave'] / wdm['n_polarisations'])
+
+    start_time = datetime.now()
+    points_x_found = get_nearest_constellation_points_unscaled(points_x_shifted, mod_type)
+    points_y_found = get_nearest_constellation_points_unscaled(points_y_shifted, mod_type)
+    print("search x and y points took",
+          (datetime.now() - start_time).total_seconds() * 1000, "ms") if verbose >= 2 else ...
+
+    start_time = datetime.now()
+    ber_x = get_ber_by_points(points_x_orig * scale_constellation, points_x_found, mod_type)
+    ber_y = get_ber_by_points(points_y_orig * scale_constellation, points_y_found, mod_type)
+    print("ber for x and y took",
+          (datetime.now() - start_time).total_seconds() * 1000, "ms") if verbose >= 2 else ...
+
+    q_x = np.sqrt(2) * sp.special.erfcinv(2 * ber_x[0])
+    q_y = np.sqrt(2) * sp.special.erfcinv(2 * ber_y[0])
+
+    # print("BER (x / y):", BER_est(wdm['m_order'], points_x_shifted, points_x_orig), BER_est(wdm['m_order'], points_y_shifted, points_y_orig))
+    print("BER (x / y):", ber_x, ber_y) if verbose >= 1 else ...
+    print("Q^2-factor (x / y):", q_x, q_y) if verbose >= 1 else ...
+
+    result = {
+        'points_x': points_x,
+        'points_x_orig': points_x_orig,
+        'points_x_shifted': points_x_shifted,
+        'points_x_found': points_x_found,
+        'points_y': points_y,
+        'points_y_orig': points_y_orig,
+        'points_y_shifted': points_y_shifted,
+        'points_y_found': points_y_found,
+        'ber_x': ber_x,
+        'ber_y': ber_y,
+        'q_x': q_x,
+        'q_y': q_y
+    }
+
     return result
\ No newline at end of file
